without warning
without trace

include machine.e
include dll.e


procedure call_func( atom func, sequence params )
	poke4( eufunc, func )
	poke4( euparam_count, length(params) )
	poke4( euparam, params )
	c_proc( eucall, {})
end procedure

--/topic wxEuphoria Developers
--/func call_member( atom fptr, atom this, object params )
--
--Calls a member function using the appropriate calling convention by platform, and returns
--an unsigned value.

--/topic wxEuphoria Developers
--/func call_member_s( atom fptr, atom this, object params )
--
--Calls a member function using the appropriate calling convention by platform, and returns
--a signed value.

--/topic wxEuphoria Developers
--/func call_back_member( integer rid, integer params )
--
--Creates a callback with the appropriate calling convention and parameters by platform.
--This is no longer needed.  You may use Euphoria's built-in call_back() function
--to create a cdecl callback:
--/code
--    my_callback = call_back( '+' & my_routine_id )
--/endcode

global function call_member( atom func, atom this, object params )
	call_func( func, this & params )
	return peek4u( euret )
end function

global function call_member_s( atom func, atom this, object params )
	call_func( func, this & params )
	return peek4s( euret )
end function

--/topic wxEuphoria Developers
--/func call_cdecl( atom func, object params )
--
global function call_cdecl( atom func, object params )
	if atom(params) then
		params = {params}
	end if
	call_func( func, params )
	return peek4u( euret )
end function

--/topic wxEuphoria Developers
--/func call_cdecl_s( atom func, object params )
--
global function call_cdecl_s( atom func, object params )
	if atom(params) then
		params = {params}
	end if
	call_func( func, params )
	return peek4s( euret )
end function

global function call_back_member( integer rid, integer params )
	return call_back( '+' & rid )
end function

sequence class_check
class_check = {{},{}}

global sequence class_table, proc_table, instance_class_table
object void

constant
CL_NAME = 1,
CL_INHERIT = 2,
CL_CONSTRUCTOR = 3,
CL_DERIVED = 4,
CL_SIZE = 5,
CL_DESTRUCTOR = 6,

PR_LIB_PROC = 1,
PR_FPTR = 2,

IC_THIS = 1,
IC_CLASS = 2

class_table = repeat( {}, 6)
proc_table = repeat( {}, 2 )
instance_class_table = repeat( {}, 2)


--/topic wxEuphoria Developers
--/func define_class(sequence name, atom hlib, atom constructor,integer bytes, atom destructor )
--
--Used to wrap a C++ class.
--<ul>
--/li /i name: the name of the class
--/li /i hlib: pointer to the dll/so of the library where the class resides
--/li /i constructor: address of the constructor to be used by /new_instance()
--/li /i bytes: size of an instance of the class
--/li /i destructor: address of the destructor to be used by /delete_instance()
--</ul>
global function define_class(sequence name,atom hlib, atom constructor,integer bytes, atom destructor )
	integer class_ix

	class_ix = find( name, class_table[CL_NAME] )
	if class_ix then
		return class_ix
	end if

	for i = 1 to length( class_table ) do
		class_table[i] = append( class_table[i], {} )
	end for

	class_ix = length( class_table[1] )
	class_table[CL_NAME][class_ix] = name
	class_table[CL_CONSTRUCTOR][class_ix] = constructor

	class_table[CL_DERIVED][class_ix] = {}
	class_table[CL_SIZE][class_ix] = bytes
	class_table[CL_DESTRUCTOR][class_ix] = destructor

	return class_ix
end function

global procedure overload_member( atom this, integer vtbl, atom new )
	
end procedure

--/topic wxEuphoria Developers
--/proc register_instance( atom this, integer class )
--
--Registers an instance of a class, so that /delete_instance() can be used to destroy it.
--This does not need to be done for objects created with /new_instance().

global procedure register_instance( atom this, integer class_ )
	integer ix
	if find( this, instance_class_table[IC_THIS] ) then
		return
	end if

	ix = find(-1, instance_class_table[IC_THIS] )
	if not ix then
		instance_class_table[IC_THIS] &= this
		instance_class_table[IC_CLASS] &= class_
	else
		instance_class_table[IC_THIS][ix] = this
		instance_class_table[IC_CLASS][ix] = class_
	end if
end procedure

--/topic Utilities
--/func lookup_class( atom this )
--
--Returns the class name for the object, as long as it was either created by /new_instance() or
--registered by /register_instance().
global function lookup_class( atom this )
	integer class_ 
	class_ = find( this, instance_class_table[IC_THIS] )
	if class_ then
		class_ = instance_class_table[IC_CLASS][class_]
	end if

	return class_
end function


--/topic Utilities
--/func lookup_class_name( sequence name )
--
--Returns the value of the associated class.
global function lookup_class_name( sequence name )
	return find( name, class_table[CL_NAME] )
end function

--/topic wxEuphoria Developers
--/func new_instance( integer class_ix, atom this, sequence params )
--
--Allocates space, calls the constructor and registers a new instance of a class.
--<ul>
--/li /i class_ix: the id returned by /define_class()
--/li /i this: pointer to pre-allocated memory, or zero if new_instance() should allocate space
--/li /i params: parameters required by the object's constructor
--</ul>
--This function should only be called by /create_func 's.  You should use /create() to 
--create new instances of objects.

global function new_instance( integer class_ix, atom this, sequence params )
	atom vtbl, pvtbl, ctor
	object void

	if not this then
		this = allocate( class_table[CL_SIZE][class_ix] )
	end if
	register_instance( this, class_ix )

	ctor = class_table[CL_CONSTRUCTOR][class_ix]

	void = call_member( ctor, this, params )

	if length(class_table[CL_DERIVED][class_ix]) then

		vtbl = peek4u(this)
		pvtbl = vtbl

		for i = 1 to length(class_table[CL_DERIVED][class_ix]) do
			poke4( pvtbl + class_table[CL_DERIVED][class_ix][i][1] * 4, class_table[CL_DERIVED][class_ix][i][2] )
		end for

		poke4(this, pvtbl )
	end if

	return this
end function


--/topic wxEuphoria Developers
--/func derived_class( sequence name, integer parent, atom hlib, sequence overrides, integer bytes )
--
--Used to derive a class from another.  This is rarely (if ever) needed in wxEuphoria.
--<ul>
--/li /i name: name of derived class
--/li /i parent: class id of parent class
--/li /i hlib: handle to dll returned from load_library
--/li /i overrides: sequence of virtual functions to override { {vtbl_entry, function callback},...}
--/li /i bytes: size, in bytes of instance of the derived class
--</ul>
global function derived_class( sequence name, integer parent, atom hlib, 
		sequence overrides, integer bytes )
	integer class_ix
	atom pvtbl

	class_ix = define_class( name, hlib, class_table[CL_CONSTRUCTOR][parent], 
		bytes, class_table[CL_DESTRUCTOR][parent] )
	class_table[CL_DERIVED][class_ix] = overrides

	return class_ix
end function

--/topic wxEuphoria Developers
--/func deref_instance( object this )
--
--Dereferences an object so that it cannot be destroyed
--by wxEuphoria.  Use this if wxWidgets takes ownership
--of an object, and will delete it automatically.
with trace
global procedure deref_instance( object this )
	integer index
	if atom(this) then
		this = {this}
	end if
	for i = 1 to length(this) do
		
		index = find( this[i], instance_class_table[IC_THIS] )
		
		if index then
			instance_class_table[IC_CLASS][index] = -1
			instance_class_table[IC_THIS][index] = -1
		end if
	end for
end procedure

--#WIN32
sequence garbage
garbage = {}
procedure collect_garbage()
	for i = 1 to length( garbage ) do
		deref_instance( garbage[i] )
		free( garbage[i] )
	end for
	garbage = {}
end procedure

global procedure free_garbage( atom ptr )
	integer ix
	ix = find( ptr, garbage )
	if ix then
		garbage = garbage[1..ix-1] & garbage[ix+1..length(garbage)]
	end if
	free( ptr )
end procedure
--#ENDWIN32

global procedure free_instance( atom ptr )
--#LINUX
--	free( ptr )
--#ENDLINUX
--#WIN32
	if not find( ptr, garbage ) then
		garbage &= ptr
	end if
	
--#ENDWIN32
end procedure

--/topic Classes
--/proc delete_instance( object this )
--
--Destroys an object and frees the memory that held it.  Can also be a sequence
--of objects to be destroyed.

global procedure delete_instance( object this )
	integer class_
	object dtor

	if atom( this ) then
		this = {this}
	end if

	for i = 1 to length(this) do
		class_ = lookup_class( this[i] )

		if class_ then
			dtor = class_table[CL_DESTRUCTOR][class_]
			if atom(dtor) then
				if dtor > 0 then
					if dtor then
						void = call_member( dtor, this[i], {} )
					end if

					free_instance( this[i] )

				elsif dtor then
					call_proc( -dtor, {this[i]})
					return
				else
					free_instance( this[i] )

				end if
			else

				free_instance( this[i] )

			end if
			deref_instance( this[i] )
		end if
	end for
--#WIN32
	collect_garbage()
--#ENDWIN32
end procedure

--/topic wxEuphoria Developers
--/proc set_destructor( integer class, integer dtor )
--
--Sets the destructor for the class to /b dtor, where /b dtor is 
--the routine id of a user supplied procedure that will be called
--instead of a wxWidgets supplied destructor.  The procedure should
--take one atom (this pointer) as its arguments, and should
--free the memory allocated by the object (because wxEuphoria won't).
--
--/code
--  ex:
--    global procedure delete_string( atom string )
--        deref_string( string )
--        free(string)
--    end procedure
--    set_destructor( wxString, routine_id("delete_string"))
--/endcode
global procedure set_destructor( integer class_, integer dtor )
	class_table[CL_DESTRUCTOR][class_] = - dtor
end procedure

--/topic wxEuphoria Developers
--/func get_destructor( integer class )
--
--If the return value is negative, then its absolute value is the
--routine id of a user supplied destructor routine.

global function get_destructor( integer class_ )
	return class_table[CL_DESTRUCTOR][class_]
end function

--/topic wxEuphoria Developers
--/func get_constructor( integer class )
-- 
--Returns the constructor for the class.

global function get_constructor( integer class_ )
	return class_table[CL_CONSTRUCTOR][class_]
end function

global function vtbl_index( atom vtbl, atom func )
	integer ix
	atom current

	current = 0
	ix = -1
	while current != func do
		ix += 1
		current = peek4u( vtbl + ix*4)
	end while
	return ix
end function



